/* pbmtoiw -- Convert a portable bitmap into Apple Imagewriter printer
	      graphics. Use: "pbmtoiw [-<hres>x<vres>] [pbmfile]". Hres is one
	      of 72, 80, 96, 107, 120, 136, 144 or 160; vres is 72 or 144. The
	      default is "-72x72". If the pbmfile argument is absent, pbmtoiw
	      reads standard input. The Imagewriter graphics are always sent to
	      standard output. */

#include <stdio.h>
#include <ctype.h>
#include <malloc.h>

/* -------------------------------------------------------------------------
   Some miscellaneous defines
   ------------------------------------------------------------------------- */

#define SETUP_FILE "/usr/local/lib/iw.setup"
#define FIELDLENGTH 10
#define FALSE 0
#define TRUE 1
#define PBM_ASCII 0
#define PBM_RAW 1

/* -------------------------------------------------------------------------
   These globals are initialized by read_setup_file()
   ------------------------------------------------------------------------- */

int paper_height, tearoff_margin, join_pages;

/* -------------------------------------------------------------------------
   These globals are initialized by read_command_line()
   ------------------------------------------------------------------------- */

FILE *input_file;
int pix_per_line, interleaved;
char esc_char;

/* -------------------------------------------------------------------------
   These globals are used to parse the "-<hres>x<vres>" option
   ------------------------------------------------------------------------- */

#define NUM_RESOLUTIONS 16
#define MAX_PIX_PER_LINE 1280

char *RES_OPTION[] =
{
	"72x72", "80x72", "96x72", "107x72", "120x72", "136x72", "144x72",
	"160x72", "72x144", "80x144", "96x144", "107x144", "120x144",
	"136x144", "144x144", "160x144"
};

int PIX_PER_LINE[] =
{
	576, 640, 768, 856, 960, 1088, 1152, 1280,
	576, 640, 768, 856, 960, 1088, 1152, 1280
};

int INTERLEAVED[] =
{
	FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,
	TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE
};

char ESC_CHAR[] = "nNEeqQpPnNEeqQpP";

/* -------------------------------------------------------------------------
   Other globals
   ------------------------------------------------------------------------- */

int image_type, image_width, image_height, original_height, bytes_per_line;
unsigned char *bitmap;
int delayed_feed = 0;


/* -------------------------------------------------------------------------
   Read paper setup parameters from file
   ------------------------------------------------------------------------- */

void read_setup_file()
{
	FILE *setup_file;
	int n;

	if ((setup_file = fopen(SETUP_FILE, "r")) == NULL)
	{
		fprintf(stderr, "pbmtoiw: can't open %s\n", SETUP_FILE);
		exit(1);
	}
	n = fscanf(setup_file, "%d %d %d", &paper_height, &tearoff_margin,
								&join_pages);
	fclose(setup_file);
	if (n != 3)
	{
		fprintf(stderr, "pbmtoiw: error reading %s\n", SETUP_FILE);
		exit(1);
	}
}

/* -------------------------------------------------------------------------
   Free education
   ------------------------------------------------------------------------- */

void syntax_abort()
{
	fputs("usage: pbmtoiw [-<hres>x<vres>] [pbmfile]\n", stderr);
	fputs("       hres = 72, 80, 96, 107, 120, 136, 144 or 160\n", stderr);
	fputs("       vres = 72 or 144\n", stderr);
	fputs("       default: hres = vres = 72\n", stderr);
	exit(1);
}

/* -------------------------------------------------------------------------
   Parse command line arguments
   ------------------------------------------------------------------------- */

void read_command_line(argc, argv)
	int argc;
	char *argv[];
{
	int i, res_opt = FALSE, res_index = 0, file_arg = 0;

	for (i = 1; i < argc; i++)
		if (argv[i][0] == '-')
		{
			if (res_opt)
				syntax_abort();
			res_index = 0;
			while (res_index < NUM_RESOLUTIONS &&
				strcmp(argv[i] + 1, RES_OPTION[res_index]))
				res_index++;
			if (res_index == NUM_RESOLUTIONS)
				syntax_abort();
			res_opt = TRUE;
		}
		else
		{
			if (file_arg)
				syntax_abort();
			file_arg = i;
		}

	pix_per_line = PIX_PER_LINE[res_index];
	interleaved = INTERLEAVED[res_index];
	esc_char = ESC_CHAR[res_index];

	if (file_arg)
	{
		input_file = fopen(argv[file_arg], "r");
		if (input_file == NULL)
		{
			fprintf(stderr, "pbmtoiw: can't open %s\n",
							argv[file_arg]);
			exit(1);
		}
	}
	else
		input_file = stdin;
}

/* -------------------------------------------------------------------------
   Read a character from the PBM stream, skipping comments
   ------------------------------------------------------------------------- */

int get_a_char()
{
	int c, comment = FALSE;

	do
	{
		if ((c = getc(input_file)) == EOF)
		{
			fputs("pbmtoiw: unexpected end of file\n", stderr);
			exit(1);
		}
		if (c == '#')
			comment = TRUE;
		else if (c == '\n' || c == '\r')
			comment = FALSE;
	}
	while (comment);
	return c;
}

/* -------------------------------------------------------------------------
   Read a whitespace-delimited string from the PBM stream
   ------------------------------------------------------------------------- */

void get_a_string(s, n)
	char *s;
	int n;
{
	int pos = 0, c;

	do
		c = get_a_char();
	while (isspace(c));
	s[pos] = c;
	do
	{
		c = get_a_char();
		if (pos < n - 1)
			s[++pos] = c;
	}
	while (!isspace(c));
	s[pos] = 0;
}

/* -------------------------------------------------------------------------
   Read a non-blank character from the PBM stream, skipping comments
   ------------------------------------------------------------------------- */

int get_a_nonspace_char()
{
	int c;

	while (isspace(c = get_a_char()));
	return c;
}

/* -------------------------------------------------------------------------
   Read PBM header
   ------------------------------------------------------------------------- */

void read_header()
{
	char field[FIELDLENGTH];

	/* Get image type */
	get_a_string(field, FIELDLENGTH);
	if (!strcmp(field, "P1"))
		image_type = PBM_ASCII;
	else if (!strcmp(field, "P4"))
		image_type = PBM_RAW;
	else
	{
		fputs("pbmtoiw: unsupported image type\n", stderr);
		exit(1);
	}
	
	/* Get image dimensions */
	get_a_string(field, FIELDLENGTH);
	image_width = atoi(field);
	get_a_string(field, FIELDLENGTH);
	original_height = atoi(field);

	/* Vertical padding to simplify printing algorithm */
	if (interleaved)
		image_height = (original_height + 15) & -16;
	else
		image_height = (original_height + 7) & -8;
}

/* -------------------------------------------------------------------------
   Allocate memory for bitmap image
   ------------------------------------------------------------------------- */

void alloc_bitmap()
{
	int bitmap_size;

	/* Lines are padded out to an even byte boundary */
	bytes_per_line = ((image_width + 15) >> 3) & -2;

	/* Allocate bitmap and initialize it to all zeros */
	bitmap_size = bytes_per_line * image_height;
	bitmap = (unsigned char *) malloc(bitmap_size);
	if (bitmap == NULL)
	{
		fputs("pbmtoiw: out of memory\n", stderr);
		exit(1);
	}
	memset(bitmap, 0, bitmap_size);
}

/* -------------------------------------------------------------------------
   Read ascii PBM file
   ------------------------------------------------------------------------- */

void read_pbm_ascii()
{
	unsigned char in_chr, cur_chr, cur_bit;
	int row, column, pos = 0;

	pos = 0;
	for (row = 0; row < original_height; row++)
	{
		cur_bit = 0x80;
		cur_chr = 0;
		for (column = 0; column < image_width; column++)
		{
			in_chr = get_a_nonspace_char();
			if (in_chr == '1')
				cur_chr |= cur_bit;
			else if (in_chr != '0')
			{
				fputs("pbmtoiw: pbm format error\n", stderr);
				exit(1);
			}
			cur_bit >>= 1;
			if (!cur_bit)
			{
				bitmap[pos++] = cur_chr;
				cur_bit = 0x80;
				cur_chr = 0;
			}
		}
		if (cur_bit != 0x80)
			bitmap[pos++] = cur_chr;
		if (pos & 1)
			pos++;
	}
}

/* -------------------------------------------------------------------------
   Read raw PBM file
   ------------------------------------------------------------------------- */

void read_pbm_raw()
{
	int row, count, pos = 0, bytes_per_pbm_line;

	bytes_per_pbm_line = (image_width + 7) >> 3;
	for (row = 0; row < original_height; row++)
	{
		count = fread(&bitmap[pos], 1, bytes_per_pbm_line, input_file);
		if (count != bytes_per_pbm_line)
		{
			fputs("pbmtoiw: pbm format error\n", stderr);
			exit(1);
		}
		pos += bytes_per_line;
	}
}

/* -------------------------------------------------------------------------
   Slurp bitmap into core
   ------------------------------------------------------------------------- */

void read_file()
{
	switch (image_type)
	{
		case PBM_ASCII:
		read_pbm_ascii();
		break;

		case PBM_RAW:
		read_pbm_raw();
		break;

		default:
		fputs("pbmtoiw: this can't happen\n", stderr);
		exit(1);
	}
}

/* -------------------------------------------------------------------------
   Perform all delayed paper feed actions
   ------------------------------------------------------------------------- */

void position_paper()
{
	int a, b;
	static int lf_pitch = 0;

	if (delayed_feed)
	{
		a = delayed_feed >> 6;
		b = delayed_feed & 63;
		if (a)
		{
			if (lf_pitch != 64)
			{
				fputs("\033T64", stdout);
				lf_pitch = 64;
			}
			while (a--)
				fputs("\r\n", stdout);
		}
		if (b)
		{
			if (lf_pitch != b)
			{
				printf("\033T%02u", b);
				lf_pitch = b;
			}
			fputs("\r\n", stdout);
		}
		delayed_feed = 0;
	}
}

/* -------------------------------------------------------------------------
   Skip n pixels vertically
   ------------------------------------------------------------------------- */

void skip_pixels(n)
	int n;
{
	delayed_feed += interleaved ? n : n << 1;
}

/* -------------------------------------------------------------------------
   Print one line of the bitmap (one pass of the print head)
   ------------------------------------------------------------------------- */

void print_line(x, y, width, feed)
	int x, y, width, feed;
{
	int out_pos, i, first_nonblank = -1, last_nonblank, num_chars;
	unsigned char bitmask, *in_char[8], out_char;
	unsigned char pixbuf[MAX_PIX_PER_LINE];

	/* Find bits to print in bitmap */
	in_char[0] = bitmap + (x >> 3) + y * bytes_per_line;
	for (i = 1; i < 8; i++)
		in_char[i] = in_char[i - 1] + (bytes_per_line << interleaved);

	/* Shuffle bits for printer output */
	bitmask = 0x80;
	for (out_pos = 0; out_pos < width; out_pos++)
	{
		out_char = 0;
		for (i = 0; i < 8; i++)
		{
			out_char >>= 1;
			if (*in_char[i] & bitmask)
				out_char |= 0x80;
		}
		pixbuf[out_pos] = out_char;
		if (out_char)
		{
			if (first_nonblank == -1)
				first_nonblank = out_pos;
			last_nonblank = out_pos;
		}
		bitmask >>= 1;
		if (!bitmask)
		{
			bitmask = 0x80;
			for (i = 0; i < 8; i++)
				in_char[i]++;
		}
	}

	/* Now print them. If there is anything to print, that is */
	if (first_nonblank == -1)
	{
		delayed_feed += feed;
		return;
	}
	position_paper();
	if (first_nonblank)
		printf("\033F%04u", first_nonblank);
	num_chars = last_nonblank - first_nonblank + 1;
	printf("\033G%04u", num_chars);
	fwrite(&pixbuf[first_nonblank], 1, num_chars, stdout);
	delayed_feed = feed;
}

/* -------------------------------------------------------------------------
   Print one page from the bitmap
   ------------------------------------------------------------------------- */

void print_page(x, y, width, height)
	int x, y, width, height;
{
	int cur_y;

	/* Print lines */
	cur_y = y;
	if (interleaved)
		while (cur_y < y + height)
		{
			print_line(x, cur_y, width, 1);
			print_line(x, cur_y + 1, width, 15);
			cur_y += 16;
		}
	else
	{
		while (cur_y < y + height)
		{
			print_line(x, cur_y, width, 16);
			cur_y += 8;
		}
	}
}

/* -------------------------------------------------------------------------
   Print the bitmap!
   ------------------------------------------------------------------------- */

void print_bitmap()
{
	int max_height, x, y, print_width, print_height;
	int image_ready, column_ready;

	max_height = (paper_height - tearoff_margin) & -8;
	tearoff_margin = paper_height - max_height;
	if (interleaved)
	{
		paper_height <<= 1;
		tearoff_margin <<= 1;
		max_height <<= 1;
	}
	else
		fputs("\033>", stdout);

	fputc('\033', stdout);
	fputc(esc_char, stdout);

	x = 0;
	print_width = pix_per_line;
	image_ready = FALSE;
	while (!image_ready)
	{
		if (x + print_width >= image_width)
		{
			print_width = image_width - x;
			image_ready = TRUE;
		}

		if (join_pages)
		{
			y = 0;
			print_page(x, y, print_width, image_height);
			skip_pixels(paper_height -
					image_height % paper_height);
		}
		else
		{
			y = 0;
			print_height = max_height;
			column_ready = FALSE;
			while (!column_ready)
			{
				if (y + print_height >= image_height)
				{
					print_height = image_height - y;
					column_ready = TRUE;
				}
				print_page(x, y, print_width, print_height);
				skip_pixels(paper_height - print_height);
				y += max_height;
			}
		}

		x += pix_per_line;
	}
	position_paper();
}

/* -------------------------------------------------------------------------
   Final cleanup
   ------------------------------------------------------------------------- */

void cleanup()
{
	free(bitmap);
	fputs("\033c", stdout);
}

/* -------------------------------------------------------------------------
   The Function Of Supreme Importance
   ------------------------------------------------------------------------- */

void main(argc, argv)
	int argc;
	char *argv[];
{
	read_setup_file();
	read_command_line(argc, argv);
	read_header();
	alloc_bitmap();
	read_file();
	print_bitmap();
	cleanup();
	exit(0);
}
